#ifndef _SOLVER_H
#define _SOLVER_H

class function
{
public:
	virtual double fun(double *w) = 0 ;
	virtual void grad(double *w, double *g) = 0 ;
	virtual void Hv(double *s, double *Hs) = 0 ;
	virtual void Xv(double *v, double *Xv) = 0;
	virtual void Xv(double *v, double *Xv, int vecs) = 0;
	virtual void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length) = 0;
	virtual double line_search(double *w, double *delta_w, double gd, double dTd, double wTd, double *f, double eta, double init_size = 1.0, double *Xd = NULL) = 0 ;

	virtual int get_nr_variable(void) = 0 ;
	virtual void get_wx(double *external_wx) = 0 ;
	virtual void get_diagH(double *M) = 0 ;
	virtual ~function(void){}
};

class SOLVER
{
public:
	SOLVER(const function *fun_obj, double eps = 0.1, int max_iter = 50000);
	~SOLVER();
	virtual void solve(double *w) = 0 ;
	void set_print_string(void (*i_print) (const char *buf));
protected:
	void info(const char *fmt,...);
	double eps;
	int max_iter;
	function *fun_obj;
	void (*solver_print_string)(const char *buf);
	int start;
	int length;
};

class MLBFGS: public SOLVER
{
public:
	MLBFGS(const function *fun_obj, int l, double eps = 0.1, int m=5, double eta = 1e-2, int max_iter = 50000);
	~MLBFGS();

	void solve(double *w);
protected:
	int M;
	int inner_product_entry(int i, int j);

	double eta;
	int l;
	int n;
	int Plength;
	int tmp_length;
	int wcounter;
	int gcounter;

	double *sqrtD_Px;

	int *displace;
	int *recv_count;
	double *recv_buffer;
private:
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* delta, double step_size);
	void compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse);
	void compute_system(double* inner_product_matrix, double* Px, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length);

};

class MLBFGS_DIAG: public MLBFGS
{
public:
	MLBFGS_DIAG(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 50000);
	~MLBFGS_DIAG();

	void solve(double *w);
private:
	int dcounter;
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* delta, double step_size);
	void compute_system(double* inner_product_matrix, double* Px, double* C_times_D, double* system, int sparse);
	void compute_system(double* inner_product_matrix, double* Px, double* C_times_D, double* system, int sparse, int *index, int index_length);
};

class LBFGS: public SOLVER
{
public:
	LBFGS(const function *fun_obj, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 100000);
	~LBFGS();

	void solve(double *w);

private:
	void update_inner_products(double** inner_product_matrix, int k, int n, int DynamicM, int previous_M, double** s, double** y, double* g, double* delta, double step_size);
	void TwoLoop(double** inner_product_matrix, double* rho, int DynamicM, int k, int n, double* delta);
	double eta;
	int M;
	int offset;
};

class TRON: public SOLVER
{
public:
	TRON(const function *fun_obj, double eps = 0.1, double default_eps = 0.01, double eps_cg = 0.1, int max_iter = 10000);
	~TRON();

	void solve(double *w);

private:
	int trpcg(double delta, double *g, double *M, double *s, double *r, bool *reach_boundary);
	double norm_inf(int n, double *x);

	double eps_cg;
	double default_eps;
};
#endif
