/*
 * solver.cpp implements the algorithms for comparison
 * In this file, we implement  LCOMM_BFGS, LCOMM_DIAG, and include the
 * implementation of TRON in multicore-LIBLINEAR
 *
 * Note that for TRON, no parallelism is achieved in the solver side except
 * for some vector-vector operations that are handled by BLAS, so this part of
 * the code is exactly the same as the single core version
 * For LCOMM, all parts except for solving the linear system using LDLT are
 * parallelized, although parts such parallelization are either in the
 * objective function class or implicitly done by optimized BLAS
 *
 * For better parallelization, optimized BLAS libraries like Intel MKL or OpenBLAS should be used
 * Our implementation here is pecifically for MKL, but change to other blas
 * libraries should be easy, and the user just need to modify the parts encompassed by
 * #ifdef MKL
 *
 * */
#include <stdio.h>
#include <stdint.h>
#include <stdarg.h>
#include <limits.h>
#include <assert.h>
#include <time.h>
#include <math.h>
#include <string.h>
#include <omp.h>
#include "solver.h"

template <class T> static inline void swap(T* a, int x, int y) { T t=a[x]; a[x]=a[y]; a[y]=t; }
#ifndef TIMER
#define TIMER
#define NS_PER_SEC 1000000000
double wall_time_diff(int64_t ed, int64_t st)
{
	return (double)(ed-st)/(double)NS_PER_SEC;
}

int64_t wall_clock_ns()
{
#if __unix__
	struct timespec tspec;
	clock_gettime(CLOCK_MONOTONIC, &tspec);
	return tspec.tv_sec*NS_PER_SEC + tspec.tv_nsec;
#else
#if __MACH__
	return 0;
#else
	struct timeval tv;
	gettimeofday( &tv, NULL );
	return tv.tv_sec*NS_PER_SEC + tv.tv_usec*1000;
#endif
#endif
}
#endif

#ifndef min
template <class T> static inline T min(T x,T y) { return (x<y)?x:y; }
#endif

#ifndef max
template <class T> static inline T max(T x,T y) { return (x>y)?x:y; }
#endif

#define MEPS 1e-16L
#define PIVOTING
#define DEBUG_LDL

// solve Ax=b , in which A[i*n+j] = Aij and we only solve submatrix A[m:m]
// This subrouine implements the LDLT algorithm with pivoting
double solve_system(const double *_A, const double *_b, int n, int m, double *x)
{
	long double *A = new long double[n * n];
	for(int i=0; i<n*n; i++)
		A[i] = _A[i];
	long double *b = new long double[n];
	for(int i=0; i<n; i++)
		b[i] = _b[i];

#ifdef PIVOTING
	int *p = new int[m];
	for(int i=0; i<m; i++)
		p[i] = i;
#endif

	// decomp
	// A -> D0
	//      L00 D1
	//      L10 L11 D2
	for(int j=0; j<m; j++){
#ifdef PIVOTING
		int pj = j;
		long double pj_val = fabsl(A[j*n+j]);
		// find piv
		for(int i=j+1; i<m; i++){
			long double pi_val = fabsl(A[i*n+i]);
			if(pj_val < pi_val){
				pj_val = pi_val;
				pj = i;
			}
		}
		// swap rows and columns of j and pj
		for(int i=0; i<m; i++)
			swap(A, j*n+i, pj*n+i);
		for(int i=0; i<m; i++)
			swap(A, i*n+j, i*n+pj);
		swap(b, j, pj);
		swap(p, j, pj);
#endif

		// loop invariant: Aik = Lik, Akk = Dk with k<j
		long double Dj = A[j*n+j];
		for(int k=0; k<j; k++){
			Dj -= A[j*n+k]*A[j*n+k] * A[k*n+k];
		}
		A[j*n+j] = Dj;
		//assert(fabs(Dj) >= MEPS);
		if(fabsl(Dj) < MEPS){
			for(int i=j+1; i<m; i++)
				A[i*n+j] = 0;
			continue; // xj should be 0
		}

		for(int i=j+1; i<m; i++){
			long double Lij = A[i*n+j];
			for(int k=0; k<j; k++){
				Lij -= A[i*n+k] * A[j*n+k] * A[k*n+k];
			}
			Lij /= Dj;
			A[i*n+j] = Lij;
		}
	}
#if 0
	printf("L\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<=i; j++)
			printf("%+.3Le ", A[i*n+j]);
		printf("\n");
	}
#endif
	// invariant: A[i*n+j] = Lij, forall i>j
	//            A[i*n+i] = Dii, forall i

	// LD y = b
	// di yi =  bi - \sum_{j<i} yj Lij
	// place y in b
	for(int i=0; i<m; i++){
		// loop invariant: bk = yk forall k<i
		long double yi = b[i];
		for(int k=0; k<i; k++)
			yi -= A[i*n+k] * b[k];
		b[i] = yi;
	}
	for(int i=0; i<m; i++){
		long double Di = A[i*n+i];
		if(fabsl(Di) < MEPS)
			b[i] = 0;
		else
			b[i] /= Di;
	}
#if 0
	printf("y\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", b[i]);
	}
#endif
	// L'x = y
	// xi = yi - \sum{j>i} xj Lij
	for(int i=m-1; i>=0; i--){
		// loop invariant: xk ready forall k>i
		long double xi = b[i];
#ifdef PIVOTING
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[p[k]];
		x[p[i]] = (double)xi;
#else
		for(int k=m-1; k>i; k--)
			xi -= A[k*n+i] * x[k];
		x[i] = (double)xi;
#endif
	}
#if 0
	printf("\n");
	for(int i=0; i<n; i++)
		printf("d %.3e\n", x[i]);

	printf("A\n");
	for(int i=0; i<n; i++){
		for(int j=0; j<n; j++)
			printf("%+.3e ", Acopy[i*n+j]);
		printf("\n");
	}
	printf("b\n");
	for(int i=0; i<n; i++){
		printf("%+.3e\n", bcopy[i]);
	}
#endif

#if 1
	double err = 0;
	for(int i=0; i<m; i++){
		double Ax_i = 0;
		for(int j=0; j<m; j++)
			Ax_i += _A[i*n+j] * x[j];
		err += fabs(Ax_i-_b[i]);
	}
	double max_pv = 0;
	for(int i=0; i<m; i++)
		if(max_pv < _A[i*n+i])
			max_pv = (double)_A[i*n+i];
	if(max_pv != 0)
		err /= max_pv;
	//printf("err=%e\n", err);
	//assert(err < 1e-5);
#endif
	delete[] A;
	delete[] b;
#ifdef PIVOTING
	delete[] p;
#endif
	return err;
}

#ifdef MKL
#include "mkl.h"
#define dnrm2_ dnrm2
#define ddot_ ddot
#define daxpy_ daxpy
#define dscal_ dscal
#define dgemv_ dgemv
#define dsyrk_ dsyrk
#define dspmv_ dspmv
#else
#ifdef __cplusplus
extern "C" {
#endif

extern double dnrm2_(int *, double *, int *);
extern double ddot_(int *, double *, int *, double *, int *);
extern int daxpy_(int *, double *, double *, int *, double *, int *);
extern int dscal_(int *, double *, double *, int *);
extern int dgemv_(char *, int *, int *, double *, double *, int *, double *, int *, double *, double *, int *);
extern int dsyrk_(char *, char *, int *, int *, double *, double *, int *, double *, double *, int *);
extern int dspmv_(char *, int *, double *, double *, double *, int *, double *, double *, int *);

#ifdef __cplusplus
}
#endif
#endif

static void default_print(const char *buf)
{
	fputs(buf,stdout);
	fflush(stdout);
}

//For the preconditioned CG in TRON
static double uTMv(int n, double *u, double *M, double *v)
{
	const int m = n-4;
	double res = 0;
	int i;
	for (i=0; i<m; i+=5)
		res += u[i]*M[i]*v[i]+u[i+1]*M[i+1]*v[i+1]+u[i+2]*M[i+2]*v[i+2]+
			u[i+3]*M[i+3]*v[i+3]+u[i+4]*M[i+4]*v[i+4];
	for (; i<n; i++)
		res += u[i]*M[i]*v[i];
	return res;
}

void SOLVER::info(const char *fmt,...)
{
	char buf[BUFSIZ];
	va_list ap;
	va_start(ap,fmt);
	vsprintf(buf,fmt,ap);
	va_end(ap);
	(*solver_print_string)(buf);
}

void SOLVER::set_print_string(void (*print_string) (const char *buf))
{
	solver_print_string = print_string;
}

SOLVER::SOLVER(const function *fun_obj, double eps, int max_iter)
{
	this->fun_obj = const_cast<function *>(fun_obj);
	this->eps = eps;
	this->max_iter = max_iter;
	solver_print_string = default_print;
}

LCOMM_BFGS::LCOMM_BFGS(const function *fun_obj, int l, double eps, int m, double eta, int threads, int max_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->M = m;
	this->l = l;
	Plength = 2 * M + 2;
	this->n = this->fun_obj->get_nr_variable();

	sqrtD_XTP = NULL;
	this->eta = eta;

	omp_set_num_threads(threads);
#ifdef MKL
	mkl_set_num_threads(threads);
#endif
}


LCOMM_BFGS::~LCOMM_BFGS()
{
	if (sqrtD_XTP != NULL)
		delete[] sqrtD_XTP;
}

//Algorithm 3 in the paper. Here A = I and b = 0, so many operations are simplified
void LCOMM_BFGS::solve(double *w)
{
	info("threads = %d\n",omp_get_max_threads());
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int inc = 1;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";

	int i;
	int *index = NULL; //Index set, used when D_w and u_w are sparse
	int index_length = 0;//the size of the index set. Set to 0 when the vectors are dense
	double f;
	int iter = 0;

	int DynamicM = 0;//min(M,iter)
	int previous_M = 0;//DynamicM at last iteration
	double gnorm;
	double step_size = 0;
	double gTp = 0;
	double err = 0;//Residual (in l1 norm) from the linear system solver
	double pTp;
	double gnorm_at_0;

	sqrtD_XTP = new double[Plength * l];

	wcounter = 0;//the column of P that stores w
	gcounter = 1;//The column of P that stores the gradient
	
	double* linear = new double[Plength];//The rhs of the linear system
	double* pTptmp = new double[Plength];//A temp vector for computing p^T p using the inner product matrix
	double *sqrt_C_times_D = new double[l];//sqrt(D_w)
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[n];//This is p
	double *step_x = new double[l];//This is X^T p
	double* t = new double[Plength];// Stores the current solution to the linear system
	double *P = new double[Plength * n];
	double *XTP = new double[Plength * l];//X^T P
	// The order is P = [w_i, g_i, w_{i+1}, g_{i+1} ...],
	// Each column represents a possible direction to update, stored in a 1d array but treated as an n * length matrix as in blas
	double *localw = P + wcounter * n;// The latest iterate (w is the final output)
	double *wx = XTP + wcounter * l; // X^T w
	double *g = P + gcounter * n;//gradient
	double *gx = XTP + gcounter * l;// X^T g
	double *system = new double[Plength * Plength];//The coefficients of the lhs of the linear system
	int sparse;//= 1 when DynamicM < M, then solve a sparse linear system by ignoring entries related to the iterates


	tmp_length = 2;//Current length of P = 2*DynamicM + 2
	double comp_sys_time=0, solve_time=0, step_time=0, update_time=0;
	double H_time = 0;
	double g_time=0, gx_time=0,line_time=0;

	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w, sizeof(double) * size_t(n));
	memset(wx, 0, sizeof(double) * size_t(l));

//Initialization: compute f(w_0), P_0, and X^T P_0
	f = fun_obj->fun(w);
	int64_t time_st = wall_clock_ns();
	fun_obj->grad(w, g);
	int64_t timer_ed = wall_clock_ns();
	int64_t iter_ed = wall_clock_ns();
	g_time = wall_time_diff(timer_ed,time_st);
	time_st = wall_clock_ns();
	fun_obj->Xv(g, gx);
	timer_ed = wall_clock_ns();
	gx_time = wall_time_diff(timer_ed,time_st);
	gnorm = dnrm2_(&n, g, &inc);
	gnorm_at_0 = gnorm;
	double tmp[2];// Compute w^T w and g^T w together using blas
	dgemv_(T, &n, &tmp_length, &one, P, &n, localw, &inc, &zero, tmp, &inc);
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;

	int64_t time_st2 = wall_clock_ns();
	double tt = wall_time_diff(time_st2,time_st2);
//Start iterating
	while (iter < max_iter)
	{
//Check stopping and print information
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);

		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed,time_st2);
		time_st2 = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm_at_0)
			break;

//Start linear system construction
		sparse = (iter <= M);// When DynamicM < M, solve a sparse linear system by ignoring entries related to the iterates
		tmp_length -= (DynamicM + 1) * sparse;//Reduced system size

		time_st = wall_clock_ns();
//LHS
		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);//Get D_w, possibly sparse
		timer_ed = wall_clock_ns();
		H_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();
		if (index_length == 0)//Dense D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse);//Compute the matrix in the linear system
		else//Sparse D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse, index, index_length);

		timer_ed = wall_clock_ns();
		comp_sys_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();
		memset(t, 0, sizeof(double) * Plength);

//Now compute the rhs vector in the linear system
		for (i=0;i<tmp_length;i++)
			linear[i] = -inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, gcounter)];//The rhs is -P^T g

//Compute t by solving the linear system
		err = solve_system(system, linear, tmp_length, tmp_length, t);

		if (sparse)//Move the coefficients to the right place for operations with P and XTP
		{
			for (i=tmp_length;i>0;i--)
			{
				t[2 * i - 1] = t[i - 1];
				t[2 * i - 2] = 0;
			}
			tmp_length = 2 * DynamicM + 2;
		}

		timer_ed = wall_clock_ns();
		solve_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

//Use t to construct information needed for line search
		gTp = 0;
		memset(pTptmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, t, &inc, &zero, pTptmp, &inc);//P^T P t
		pTp = ddot_(&tmp_length, t, &inc, pTptmp, &inc);
		gTp = pTptmp[gcounter];
		double wTp = pTptmp[wcounter];

		if (gTp > 0)// If the update direction is nondescent: reverse the direction
		{
			gTp = -gTp;
			wTp = -wTp;
			dscal_(&tmp_length, &minusone, t, &inc);
		}

//If g and p are almost orthogonal, then do gradient descent instead
		if (fabs(gTp) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(t, 0, sizeof(double) * Plength);
			t[gcounter] = -1;
			pTp = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			gTp = -pTp;
			wTp  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
#pragma omp parallel for private(i) schedule(static) if(n > 800)
			for (i=0;i<n;i++)
				step[i] = 0;
#pragma omp parallel for private(i) schedule(static) if(l > 800)
			for (i=0;i<l;i++)
				step_x[i] = 0;
			daxpy_(&n, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else//Otherwise, compute Pt and (X^T P) t
		{
			dgemv_(N, &n, &tmp_length, &one, P, &n, t, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, XTP, &l, t, &inc, &zero, step_x, &inc);
		}
		timer_ed = wall_clock_ns();
		step_time += wall_time_diff(timer_ed,time_st);

//Line search
		double oldf = f;
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st2);
		pTp = ddot_(&n, step, &inc, step, &inc);
		wTp = ddot_(&n, localw, &inc, step, &inc);
		time_st2 = wall_clock_ns();
		time_st = wall_clock_ns();
		step_size = fun_obj->line_search(step, gTp, pTp, wTp, &f, eta, 1.0, step_x);
		timer_ed = wall_clock_ns();
		line_time += wall_time_diff(timer_ed,time_st);
		time_st = wall_clock_ns();
		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}

//Update w, P, X^T P, and P^T P
		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 2 * DynamicM + 2;

		//Pointer to the right place of w
		//Circular storage to avoid moving memory
		wcounter = (wcounter + 2) % Plength;
		gcounter = wcounter + 1;
//First update w, X^T w and f(w)
		if (step_size == 1)//Separated operation when stepsize = 1 for better numerical stability and efficiency
		{
			double *tmp_ptr = P + wcounter * n;
#pragma omp parallel for private(i) schedule(static) if(n>800)
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * n;
		}
		else
		{
			double *tmp_ptr = P + wcounter * n;
#pragma omp parallel for private(i) schedule(static) if(n>800)
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step_size * step[i];
			localw = P + wcounter * n;
		}
		wx = XTP + wcounter * l;
		iter_ed = wall_clock_ns();
		update_time += wall_time_diff(timer_ed,time_st);
		tt += wall_time_diff(iter_ed, time_st2);
		f = fun_obj->fun(localw);
		time_st = wall_clock_ns();
		time_st2 = wall_clock_ns();
		double actred = oldf - f;
		//numerical error handling
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}
		if (fabs(actred) <= 1.0e-30*fabs(f))
		{
			info("WARNING: actred too small\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}

		fun_obj->get_wx(wx);

//Update P and X^T P
		g = P + gcounter * n;
		gx = XTP + gcounter * l;
		time_st = wall_clock_ns();
		fun_obj->grad(localw, g);
		timer_ed = wall_clock_ns();
		g_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();
		fun_obj->Xv(g, gx);
		timer_ed = wall_clock_ns();
		gx_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

//Update P^T P
		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, t, step_size);

		timer_ed = wall_clock_ns();
		update_time += wall_time_diff(timer_ed,time_st);
	}
	memcpy(w, localw, sizeof(double) * n);

	delete[] pTptmp;
	delete[] t;
	delete[] system;
	delete[] XTP;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
}

void LCOMM_BFGS::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	char U[] = "U";
	char N[] = "N";

	int i,j;

	//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P
#pragma omp parallel for private(i,j) schedule(static)
	for (j=0;j<l;j++)
	{
		double *ptr = sqrtD_XTP + j * tmp_length;
		for (i=0;i<tmp_length;i++)
			ptr[i] = sqrt_C_times_D[j] * XTP[(i + (i + 1) * sparse) * l + j];
	}

	//Compute the part of the linear system from P^T P
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, j + (j + 1) * sparse)];
	dsyrk_(U, N, &tmp_length, &l, &one, sqrtD_XTP, &tmp_length, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[j * tmp_length + i] = system[i * tmp_length + j];//Update the lower triangular part by copy-pasting
}

void LCOMM_BFGS::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double one = 1.0;
	char U[] = "U";
	char N[] = "N";

	int i,j;
	//Fill in the entries of sqrt(D) X^T P corresponds to the i-th column of X^T P, ignore columns from X^T P corresponding to zero D_w entries
#pragma omp parallel for private(i,j) schedule(static)
	for (j=0;j<index_length;j++)
	{
		double *ptr = sqrtD_XTP + j * tmp_length;
		for (i=0;i<tmp_length;i++)
			ptr[i] = sqrt_C_times_D[j] * XTP[(i + (i + 1) * sparse) * l + index[j]];
	}

	//Compute the part of the linear system from P^T P
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i + (i + 1) * sparse, j + (j + 1) * sparse)];

	dsyrk_(U, N, &tmp_length, &l, &one, sqrtD_XTP, &tmp_length, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part, and note that sqrtD_XTP has only index_length rows
	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[j * tmp_length + i] = system[i * tmp_length + j];
}
//
//If i < j then access the (j,i) entry instead as P^T P is triangular
inline int LCOMM_BFGS::inner_product_entry(int i, int j)
{
	int index1, index2;
	index1 = (i > j) * i + (i <= j) * j;
	index2 = i + j - index1;
	return index1 * (index1 + 1) / 2 + index2;
}

void LCOMM_BFGS::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size)
{
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";
	int inc = 1;

	int i;
	int prev_length = 2 * previous_M + 2;//#columns of P in last iteration
	double *PTg = new double[tmp_length];
	double *PTw = new double[tmp_length];
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + gcounter * n, &inc, &zero, PTg, &inc);//PTg = P^T g

	int prev_w = wcounter - 2;//Place of last iterate
	if (wcounter == 0)
		prev_w = 2 * DynamicM;//circular storage

	if (step_size != 1)
		dscal_(&prev_length, &step_size, t, &inc);//Incorporate stepsize into t to reduce calculation
	t[prev_w] += 1;//Now w_new = w_old + step_size*p = Pt

	dspmv_(U, &prev_length, &one, inner_product_matrix, t, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, t, &inc);

#pragma omp parallel for private(i) schedule(static) if(tmp_length > 800)
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
#pragma omp parallel for private(i) schedule(static) if(tmp_length > 800)
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];


	delete[] PTg;
	delete[] PTw;
}

LCOMM_DIAG::LCOMM_DIAG(const function *fun_obj, int l, double eps, int m, double eta, int threads, int max_iter): LCOMM_BFGS(fun_obj, l, eps, m, eta, threads, max_iter)
{
	this->M = m;//M is #historical iterations
	Plength = 3 * M + 3;//3 vectors per iteration in P
	this->n = this->fun_obj->get_nr_variable();
}

LCOMM_DIAG::~LCOMM_DIAG()
{
}

//Algorithm 3 in the paper. Here A = I and b = 0, so many operations are simplified
void LCOMM_DIAG::solve(double *w)
{
	info("threads = %d\n",omp_get_max_threads());
	double one = 1.0;
	double minusone = -1.0;
	double zero = 0.0;
	int iter = 0, inc = 1;
	int two = 2;
	char N[] = "N";
	char U[] = "U";
	char T[] = "T";
	sqrtD_XTP = new double[Plength * l];
	int i;
	double f;
	double gnorm;
	double step_size = 0;
	double gTp = 0;
	double err = 0;
	double pTp;
	double gnorm_at_0;
	double tt = 0;//Timer
	int *index = NULL; //Index set, used when D_w and u_w are sparse
	int index_length = 0;//the size of the index set. Set to 0 when the vectors are dense

	int DynamicM = 0;//min(M,iter)
	int previous_M = 0;//DynamicM at last iteration

	wcounter = 0;//the column of P that stores w
	gcounter = 1;//The column of P that stores the gradient
	dcounter = 2;//The column of P that stores the current D^{-1} g
	
	double* linear = new double[Plength];//The rhs of the linear system
	double* pTptmp = new double[Plength];//A temp vector for computing p^T p using the inner product matrix
	double *sqrt_C_times_D = new double[l];//sqrt(D_w)
	double *inner_product_matrix = new double[Plength * Plength];//symmetric matrix, so only need to store the upper triangular part
	double *step = new double[n];//This is p
	double *step_x = new double[l];//This is X^T p
	double* t = new double[Plength];// Stores the current solution to the linear system
	double *P = new double[Plength * n];
	double *XTP = new double[Plength * l];//X^T P
	// The order is P = [w_i, g_i, D_i^{-1} g_i, w_{i+1}, g_{i+1}, D_{i+1}^{-1} g_{i+1} ...,
	double *localw = P + wcounter * n;// The latest iterate (w is the final output)
	double *wx = XTP + wcounter * l; // X^T w
	double *g = P + gcounter * n;//gradient
	double *gx = XTP + gcounter * l;// X^T g
	double *D_inverse_g = P + dcounter * n;
	double *D_inverse_g_x = XTP + dcounter * l;
	double *system = new double[Plength * Plength];//The coefficients of the lhs of the linear system
	int sparse;//= 1 when DynamicM < M, then solve a sparse linear system by ignoring entries related to the iterates
	tmp_length = 3;//Current length of P

	double comp_sys_time=0, solve_time=0, step_time=0, update_time=0;
	double H_time = 0;
	double g_time=0, gx_time=0,line_time=0;
	memset(w, 0, sizeof(double) * size_t(n));
	memcpy(localw, w, sizeof(double) * size_t(n));
	memset(wx, 0, sizeof(double) * size_t(l));

//Initialization: compute f(w_0), P_0, and X^T P_0
	f = fun_obj->fun(w);
	int64_t time_st = wall_clock_ns();
	fun_obj->grad(w, g);
	fun_obj->get_diagH(D_inverse_g);//First store D
#pragma omp parallel for private(i) schedule(static) if (n>800)
	for (i=0;i<n;i++)
		D_inverse_g[i] = g[i] / D_inverse_g[i];//Change to g/D
	int64_t timer_ed = wall_clock_ns();
	int64_t iter_ed = wall_clock_ns();
	g_time = wall_time_diff(timer_ed,time_st);
	time_st = wall_clock_ns();
	fun_obj->Xv(g, gx, 2);//D_inverse_g is right after g, so compute together
	timer_ed = wall_clock_ns();
	gx_time = wall_time_diff(timer_ed,time_st);
	gnorm = dnrm2_(&n, g, &inc);
	gnorm_at_0 = gnorm;// For stopping condition

	double tmp[5];//Compute the initial inner product entries using blas so need a buffer

	dgemv_(T, &n, &tmp_length, &one, P, &n, P + dcounter * n, &inc, &zero, tmp, &inc);//P^T (inv(D)*g) in the first 3 entries of tmp
	dgemv_(T, &n, &two, &one, P, &n, P + wcounter * n, &inc, &zero, tmp + 3, &inc);//P^T w in the last 2 entries (ignore the repeated computation for w and D^{-1} g
	inner_product_matrix[inner_product_entry(dcounter, wcounter)] = tmp[wcounter];
	inner_product_matrix[inner_product_entry(dcounter, gcounter)] = tmp[gcounter];
	inner_product_matrix[inner_product_entry(dcounter, dcounter)] = tmp[dcounter];
	inner_product_matrix[inner_product_entry(wcounter, wcounter)] = tmp[wcounter + 3];
	inner_product_matrix[inner_product_entry(wcounter, gcounter)] = tmp[gcounter + 3];
	inner_product_matrix[inner_product_entry(gcounter, gcounter)] = gnorm * gnorm;

	int64_t time_st2 = wall_clock_ns();
	tt = wall_time_diff(time_st2,time_st2);
//Start iterating
	while (iter < max_iter)
	{
//Check stopping and print information
		gnorm = sqrt(inner_product_matrix[inner_product_entry(gcounter, gcounter)]);//the gradient norm squared is an entry in the inner product matrix

		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed,time_st2);
		time_st2 = wall_clock_ns();
		info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
		if (gnorm <= eps*gnorm_at_0)
			break;

//Start linear system construction
		sparse = (iter <= M);// When DynamicM < M, solve a sparse linear system by ignoring entries related to the iterates
		tmp_length -= (DynamicM + 1) * sparse;//Reduced system size

//LHS
		time_st = wall_clock_ns();
		fun_obj->sqrt_Hessian_coeff(sqrt_C_times_D, &index, &index_length);//Get D_w, possibly sparse
		timer_ed = wall_clock_ns();
		H_time += wall_time_diff(timer_ed,time_st);
		if (index_length == 0)//Dense D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse);//Compute the matrix in the linear system
		else//Sparse D_w
			compute_system(inner_product_matrix, XTP, sqrt_C_times_D, system, sparse, index, index_length);
		timer_ed = wall_clock_ns();
		comp_sys_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

//Now compute the rhs vector in the linear system
		if (sparse)
			for (i=0;i<tmp_length / 2;i++)
			{
				linear[2 * i] = -inner_product_matrix[inner_product_entry(3 * i + 1, gcounter)];
				linear[2 * i + 1] = -inner_product_matrix[inner_product_entry(3 * i + 2, gcounter)];
			}
		else
			for (i=0;i<tmp_length;i++)
				linear[i] = -inner_product_matrix[inner_product_entry(i, gcounter)];

//Compute t by solving the linear system
		err = solve_system(system, linear, tmp_length, tmp_length, t);

		if (sparse)//Move the coefficients to the right place for operations with P and XTP
		{
			for (i=tmp_length / 2;i>0;i--)
			{
				t[3 * i - 1] = t[2 * i - 1];
				t[3 * i - 2] = t[2 * i - 2];
				t[3 * i - 3] = 0;
			}
			tmp_length = 3 * DynamicM + 3;
		}
		timer_ed = wall_clock_ns();
		solve_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

//Use t to construct information needed for line search
		gTp = 0;
		memset(pTptmp, 0, sizeof(double) * Plength);
		dspmv_(U, &tmp_length, &one, inner_product_matrix, t, &inc, &zero, pTptmp, &inc);//P^T P t
		pTp = ddot_(&tmp_length, t, &inc, pTptmp, &inc);
		gTp = pTptmp[gcounter];
		double wTp = pTptmp[wcounter];

		if (gTp > 0)// If the update direction is nondescent: reverse the direction
		{
			gTp = -gTp;
			wTp = -wTp;
			dscal_(&tmp_length, &minusone, t, &inc);
		}
		
//If g and p are almost orthogonal, then do gradient descent instead
		if (fabs(gTp) < inner_product_matrix[inner_product_entry(gcounter, gcounter)] * 1e-12)
		{
			memset(t, 0, sizeof(double) * Plength);
			t[gcounter] = -1;
			pTp = inner_product_matrix[inner_product_entry(gcounter, gcounter)];
			gTp = -pTp;
			wTp  = -inner_product_matrix[inner_product_entry(wcounter, gcounter)];
#pragma omp parallel for private(i) schedule(static) if (n > 800)
			for (i=0;i<n;i++)
				step[i] = 0;
#pragma omp parallel for private(i) schedule(static) if (l > 800)
			for (i=0;i<l;i++)
				step_x[i] = 0;
			daxpy_(&n, &minusone, g, &inc, step, &inc);
			daxpy_(&l, &minusone, gx, &inc, step_x, &inc);
		}
		else//Otherwise, compute Pt and (X^T P) t
		{
			dgemv_(N, &n, &tmp_length, &one, P, &n, t, &inc, &zero, step, &inc);
			dgemv_(N, &l, &tmp_length, &one, XTP, &l, t, &inc, &zero, step_x, &inc);
		}
		timer_ed = wall_clock_ns();
		step_time += wall_time_diff(timer_ed,time_st);

//Line search
		double oldf = f;
		iter_ed = wall_clock_ns();
		tt += wall_time_diff(iter_ed, time_st2);
		pTp = ddot_(&n, step, &inc, step, &inc);
		wTp = ddot_(&n, localw, &inc, step, &inc);
		time_st2 = wall_clock_ns();
		time_st = wall_clock_ns();
		step_size = fun_obj->line_search(step, gTp, pTp, wTp, &f, eta, 1.0, step_x);
		timer_ed = wall_clock_ns();
		line_time += wall_time_diff(timer_ed,time_st);
		time_st = wall_clock_ns();

		if (step_size == 0)
		{
			info("WARNING: step size too small\n");
			break;
		}
//Update w, P, X^T P, and P^T P
		iter++;
		previous_M = DynamicM;
		DynamicM = min(iter, M);
		tmp_length = 3 * DynamicM + 3;

		//Pointer to the right place of w
		//Circular storage to avoid moving memory
		wcounter = (wcounter + 3) % Plength;
		gcounter = wcounter + 1;
		dcounter = wcounter + 2;

//First update w, X^T w and f(w)
		if (step_size == 1)//Separated operation when stepsize = 1 for better numerical stability and efficiency
		{
			double *tmp_ptr = P + wcounter * n;
#pragma omp parallel for private(i) schedule(static) if (n > 800)
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step[i];
			localw = P + wcounter * n;
		}
		else
		{
			double *tmp_ptr = P + wcounter * n;
#pragma omp parallel for private(i) schedule(static) if (n > 800)
			for (i=0;i<n;i++)
				tmp_ptr[i] = localw[i] + step_size * step[i];
			localw = P + wcounter * n;
		}
		wx = XTP + wcounter * l;

		iter_ed = wall_clock_ns();
		update_time += wall_time_diff(timer_ed,time_st);
		tt += wall_time_diff(iter_ed, time_st2);

		f = fun_obj->fun(localw);
		time_st = wall_clock_ns();
		time_st2 = wall_clock_ns();
		double actred = oldf - f;
		//numerical error handling
		if (fabs(actred) < 0)
		{
			info("WARNING: actred < 0\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}
		if (fabs(actred) <= 1.0e-20*fabs(f))
		{
			info("WARNING: actred too small\n");
			info("iter %d m %d |g| %5.3e f %15.20e step %5.3e err %5.3e time %5.3e\n", iter,DynamicM, gnorm, f, step_size, err, tt);
			break;
		}

		fun_obj->get_wx(wx);

//Update P and X^T P
		g = P + gcounter * n;
		gx = XTP + gcounter * l;
		time_st = wall_clock_ns();
		fun_obj->grad(localw, g);

		D_inverse_g = P + dcounter * n;
		D_inverse_g_x = XTP + dcounter * l;
		fun_obj->get_diagH(D_inverse_g);//First store D
#pragma omp parallel for private(i) schedule(static) if (n > 800)
		for (i=0;i<n;i++)
			D_inverse_g[i] = g[i] / D_inverse_g[i];//Change to g/D
		timer_ed = wall_clock_ns();
		g_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

		fun_obj->Xv(g, gx, 2);//D_inverse_g is right after g, so compute together

		timer_ed = wall_clock_ns();
		gx_time += wall_time_diff(timer_ed,time_st);

		time_st = wall_clock_ns();

//Update P^T P
		update_inner_products(inner_product_matrix, DynamicM, previous_M, P, t, step_size);
		timer_ed = wall_clock_ns();
		update_time += wall_time_diff(timer_ed,time_st);
	}
	memcpy(w, localw, sizeof(double) * n);

	delete[] pTptmp;
	delete[] t;
	delete[] system;
	delete[] XTP;
	delete[] P;
	delete[] step;
	delete[] step_x;
	delete[] sqrt_C_times_D;
	delete[] linear;
	delete[] inner_product_matrix;
}

void LCOMM_DIAG::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse)
{
	double one = 1.0;
	char U[] = "U";
	char N[] = "N";

	int i,j;

	if (sparse)
#pragma omp parallel for private(i,j) schedule(static)
		for (j=0;j<l;j++)
		{
			double *ptr = sqrtD_XTP + j * tmp_length;
			for (i=0;i<tmp_length/2;i++)//Only skip columns = 3j for the iterates
			{
				ptr[2 * i] = sqrt_C_times_D[j] * XTP[(3 * i + 1) * l + j];
				ptr[2 * i + 1] = sqrt_C_times_D[j] * XTP[(3 * i + 2) * l + j];
			}
		}
	else
#pragma omp parallel for private(i,j) schedule(static)
		for (j=0;j<l;j++)
		{
			double *ptr = sqrtD_XTP + j * tmp_length;
			for (i=0;i<tmp_length;i++)
				ptr[i] = sqrt_C_times_D[j] * XTP[i * l + j];
		}

	if (sparse)
	{
		//The following 2 fors compute the part of the linear system from P^T P
		for (i=0;i<tmp_length/2;i++)
		{
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];

			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	}
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i,j)];

	dsyrk_(U, N, &tmp_length, &l, &one, sqrtD_XTP, &tmp_length, &one, system, &tmp_length);//system += (sqrtD_XTP)^T sqrtD_XTP, but only the upper triangular part

	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[j * tmp_length + i] = system[i * tmp_length + j];//Update the lower triangular part by copy-pasting
}


void LCOMM_DIAG::compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length)
{
	double one = 1.0;
	char U[] = "U";
	char N[] = "N";

	int i,j;
	if (sparse)
#pragma omp parallel for private(i,j) schedule(static)
		for (j=0;j<index_length;j++)
		{
			double *ptr = sqrtD_XTP + j * tmp_length;
			for (i=0;i<tmp_length/2;i++)
			{
				ptr[2 * i] = sqrt_C_times_D[j] * XTP[(3 * i + 1) * l + index[j]];
				ptr[2 * i + 1] = sqrt_C_times_D[j] * XTP[(3 * i + 2) * l + index[j]];
			}
		}
	else
#pragma omp parallel for private(i,j) schedule(static)
		for (j=0;j<index_length;j++)
		{
			double *ptr = sqrtD_XTP + j * tmp_length;
			for (i=0;i<tmp_length;i++)
				ptr[i] = sqrt_C_times_D[j] * XTP[i * l + index[j]];
		}

	if (sparse)
	{
		for (i=0;i<tmp_length/2;i++)
		{
			for (j=0;j < i;j++)
			{
				system[2 * i * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 1)];
				system[2 * i * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * j + 2)];
			}
			system[2 * i * tmp_length + 2 * i] = inner_product_matrix[inner_product_entry(3 * i + 1, 3 * i + 1)];

			for (j=0;j <= i;j++)
			{
				system[(2 * i + 1) * tmp_length + 2 * j] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 1)];
				system[(2 * i + 1) * tmp_length + 2 * j + 1] = inner_product_matrix[inner_product_entry(3 * i + 2, 3 * j + 2)];
			}
		}
	}
	else
		for (i=0;i<tmp_length;i++)
			for (j=0;j<=i;j++)
				system[i * tmp_length + j] = inner_product_matrix[inner_product_entry(i,j)];

	dsyrk_(U, N, &tmp_length, &index_length, &one, sqrtD_XTP, &tmp_length, &one, system, &tmp_length);

	for (i=0;i<tmp_length;i++)
		for (j=0;j<=i;j++)
			system[j * tmp_length + i] = system[i * tmp_length + j];
}


void LCOMM_DIAG::update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size)
{
	int inc = 1;
	double one = 1.0;
	double zero = 0;
	char U[] = "U";
	char T[] = "T";

	int i;
	int prev_length = 3 * previous_M + 3;//#columns of P in last iteration
	double *PTg = new double[tmp_length * 2];
	double *PTd = PTg + tmp_length;//d stands for inv(D)*g
	double *PTw = new double[tmp_length];
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + gcounter * n, &inc, &zero, PTg, &inc);//P^T g
	dgemv_(T, &n, &tmp_length, &one, P, &n, P + dcounter * n, &inc, &zero, PTd, &inc);//P^T d

	int prev_w = wcounter - 3;//Place of last iterate
	if (wcounter == 0)
		prev_w = 3 * DynamicM;//circular storage

	if (step_size != 1)
		dscal_(&prev_length, &step_size, t, &inc);//Incorporate stepsize into t to reduce calculation
	t[prev_w] += 1;//Now w_new = w_old + step_size*p = Pt

	dspmv_(U, &prev_length, &one, inner_product_matrix, t, &inc, &zero, PTw, &inc);
	PTw[wcounter] = ddot_(&prev_length, PTw, &inc, t, &inc);

#pragma omp parallel for private(i) schedule(static) if (tmp_length > 800)
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, wcounter)] = PTw[i];
#pragma omp parallel for private(i) schedule(static) if (tmp_length > 800)
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, gcounter)] = PTg[i];
#pragma omp parallel for private(i) schedule(static) if (tmp_length > 800)
	for (i=0;i<tmp_length;i++)
		inner_product_matrix[inner_product_entry(i, dcounter)] = PTd[i];

	delete[] PTw;
	delete[] PTg;
}

TRON::TRON(const function *fun_obj, double eps, double default_eps, double eps_cg, int max_iter, int max_total_cg_iter): SOLVER(fun_obj, eps, max_iter)
{
	this->eps_cg=eps_cg;
	this->default_eps = default_eps;
	this->max_total_cg_iter = max_total_cg_iter;
}

TRON::~TRON()
{
}

void TRON::solve(double *w)
{
	// Parameters for updating the iterates.
	double eta0 = 1e-4, eta1 = 0.25, eta2 = 0.75;

	// Parameters for updating the trust region size delta.
	double sigma1 = 0.25, sigma2 = 0.5, sigma3 = 4;

	int n = fun_obj->get_nr_variable();
	int i, cg_iter;
	double delta=0, sMnorm, one=1.0;
	double alpha, f, fnew, prered, actred, gs;
	int search = 1, iter = 1, inc = 1;
	double *s = new double[n];
	double *r = new double[n];
	double *g = new double[n];

	const double alpha_pcg = 0.01;
	double *M = new double[n];

	// calculate gradient norm at w=0 for stopping condition.
	memset(w, 0, sizeof(double) * size_t(n));

	int64_t time_st = wall_clock_ns();
	f = fun_obj->fun(w);
	fun_obj->grad(w, g);
	delta = dnrm2_(&n, g, &inc);
	double gnorm = delta;
	double gnorm0 = gnorm;

	if (gnorm <= eps*gnorm0)
		search = 0;

	iter = 0;
	int64_t iter_ed = wall_clock_ns();
	double tt = wall_time_diff(iter_ed, time_st);
	info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, 0.0, 0.0, delta, f, gnorm, 0, tt);
	time_st = wall_clock_ns();

	double *w_new = new double[n];
	bool reach_boundary;
	int accum_cg = 0;
	while (iter <= max_iter && search)
	{
		if (accum_cg >= max_total_cg_iter)
			break;
		fun_obj->get_diagH(M);
		for(i=0; i<n; i++)
			M[i] = (1-alpha_pcg) + alpha_pcg*M[i];
		if (iter == 1)
			delta = sqrt(uTMv(n, g, M, g));
		cg_iter = trpcg(delta, g, M, s, r, &reach_boundary);
		accum_cg += cg_iter;

		memcpy(w_new, w, sizeof(double) * size_t(n));
		daxpy_(&n, &one, s, &inc, w_new, &inc);

		gs = ddot_(&n, g, &inc, s, &inc);
		prered = -0.5*(gs-ddot_(&n, s, &inc, r, &inc));
		fnew = fun_obj->fun(w_new);

		// Compute the actual reduction.
		actred = f - fnew;

		// On the first iteration, adjust the initial step bound.
		sMnorm = sqrt(uTMv(n, s, M, s));
		if (iter == 1)
			delta = min(delta, sMnorm);

		// Compute prediction alpha*snorm of the step.
		if (fnew - f - gs <= 0)
			alpha = sigma3;
		else
			alpha = max(sigma1, -0.5*(gs/(fnew - f - gs)));

		// Update the trust region bound according to the ratio of actual to predicted reduction.
		if (actred < eta0*prered)
			delta = min(alpha*sMnorm, sigma2*delta);
		else if (actred < eta1*prered)
			delta = max(sigma1*delta, min(alpha*sMnorm, sigma2*delta));
		else if (actred < eta2*prered)
			delta = max(sigma1*delta, min(alpha*sMnorm, sigma3*delta));
		else
		{
			if (reach_boundary)
				delta = sigma3*delta;
			else
				delta = max(delta, min(alpha*sMnorm, sigma3*delta));
		}

		if (actred > eta0*prered)
		{
			iter++;
			memcpy(w, w_new, sizeof(double) * size_t(n));
			f = fnew;
			fun_obj->grad(w, g);

			gnorm = dnrm2_(&n, g, &inc);
			iter_ed = wall_clock_ns();
			tt += wall_time_diff(iter_ed, time_st);
			info("iter %2d act %5.3e pre %5.3e delta %5.3e f %15.20e |g| %5.3e CG %3d time %5.3e\n", iter, actred, prered, delta, f, gnorm, cg_iter, tt);
			time_st = wall_clock_ns();


			if (gnorm <= eps*gnorm0)
				break;
		}
		if (f < -1.0e+32)
		{
			info("WARNING: f < -1.0e+32\n");
			break;
		}
		if (prered <= 0)
		{
			info("WARNING: prered <= 0\n");
			break;
		}
		if (fabs(actred) <= 1.0e-12*fabs(f) &&
			fabs(prered) <= 1.0e-12*fabs(f))
		{
			info("WARNING: actred and prered too small\n");
			break;
		}
	}

	delete[] g;
	delete[] r;
	delete[] w_new;
	delete[] s;
}

int TRON::trpcg(double delta, double *g, double *M, double *s, double *r, bool *reach_boundary)
{
	int i, inc = 1;
	int n = fun_obj->get_nr_variable();
	double one = 1;
	double *d = new double[n];
	double *Hd = new double[n];
	double zTr, znewTrnew, alpha, beta, cgtol;
	double *z = new double[n];

	*reach_boundary = false;
	for (i=0; i<n; i++)
	{
		s[i] = 0;
		r[i] = -g[i];
		z[i] = r[i] / M[i];
		d[i] = z[i];
	}

	zTr = ddot_(&n, z, &inc, r, &inc);
	cgtol = eps_cg*sqrt(zTr);
	int cg_iter = 0;

	while (1)
	{
		if (sqrt(zTr) <= cgtol)
			break;
		cg_iter++;
		fun_obj->Hv(d, Hd);

		alpha = zTr/ddot_(&n, d, &inc, Hd, &inc);
		daxpy_(&n, &alpha, d, &inc, s, &inc);

		double sMnorm = sqrt(uTMv(n, s, M, s));
		if (sMnorm > delta)
		{
			info("cg reaches trust region boundary\n");
			*reach_boundary = true;
			alpha = -alpha;
			daxpy_(&n, &alpha, d, &inc, s, &inc);

			double sTMd = uTMv(n, s, M, d);
			double sTMs = uTMv(n, s, M, s);
			double dTMd = uTMv(n, d, M, d);
			double dsq = delta*delta;
			double rad = sqrt(sTMd*sTMd + dTMd*(dsq-sTMs));
			if (sTMd >= 0)
				alpha = (dsq - sTMs)/(sTMd + rad);
			else
				alpha = (rad - sTMd)/dTMd;
			daxpy_(&n, &alpha, d, &inc, s, &inc);
			alpha = -alpha;
			daxpy_(&n, &alpha, Hd, &inc, r, &inc);
			break;
		}
		alpha = -alpha;
		daxpy_(&n, &alpha, Hd, &inc, r, &inc);

		for (i=0; i<n; i++)
			z[i] = r[i] / M[i];
		znewTrnew = ddot_(&n, z, &inc, r, &inc);
		beta = znewTrnew/zTr;
		dscal_(&n, &beta, d, &inc);
		daxpy_(&n, &one, z, &inc, d, &inc);
		zTr = znewTrnew;
	}

	delete[] d;
	delete[] Hd;
	delete[] z;

	return(cg_iter);
}


double TRON::norm_inf(int n, double *x)
{
	double dmax = fabs(x[0]);
	for (int i=1; i<n; i++)
		if (fabs(x[i]) >= dmax)
			dmax = fabs(x[i]);
	return(dmax);
}
