#ifndef _SOLVER_H
#define _SOLVER_H

class function
{
public:
	virtual double fun(double *w) = 0 ; //Evaluate f(w)
	virtual void grad(double *w, double *g) = 0 ;
	virtual void Xv(double *v, double *Xv) = 0; //Xv = X*v
	virtual void Xv(double *v, double *Xv, int vecs) = 0; //Xv with v being a vector of n * vecs. Used in LCOMM_DIAG
	virtual void sqrt_Hessian_coeff(double *sqrt_C_times_D, int **index, int *index_length) = 0; //Compute the the square root of the diagonal matrix D_w in Sec 5.2 of the paper, stored in vector form
	virtual void grad_coeff(double *C_times_g) = 0;// Compute the u_w vector in Sec 5.2.
	virtual double compute_rho() = 0; // Compute the Lipschitz constant for the gradient
	virtual double line_search(double *p, double gTp, double pTp, double wTp, double *f, double eta, double init_size = 1.0, double *Xp = NULL) = 0 ;// w is the current iterate, p the update direction, g the gradient at w, Xd the vector storing X^T d; on return, fnew is the new objective value and the return value is the step size, but w is not updated

	virtual int get_nr_variable(void) = 0 ;
	virtual void get_wx(double *external_wx) = 0 ;
	virtual void get_diagH(double *M) = 0 ;
	virtual ~function(void){}
};

class SOLVER //Abstract class for solvers
{
public:
	SOLVER(const function *fun_obj, double eps = 0.1, int max_iter = 500000);// eps for the stopping condition and max_iter is the max number of iterations allowed
	virtual void solve(double *w) = 0 ;//Solve the optimization problem
	void set_print_string(void (*i_print) (const char *buf));//Either output messages to stdout or don't output any message
	virtual ~SOLVER(void){}
protected:
	void info(const char *fmt,...);//Function for flush output to stdout
	double eps;
	int max_iter;
	function *fun_obj;//objective function handle
	void (*solver_print_string)(const char *buf);
};


class LCOMM_OPT: public SOLVER
{
public:
	LCOMM_OPT(const function *fun_obj, int l, double eps = 0.1, double eta = 1e-2, int multi_iter = 0, const double sigma = 1.0, int max_iter = 500000);
	~LCOMM_OPT();

	void solve(double *w);

protected:
	int inner_product_entry(int i, int j);//Access The (i,j) entry of V_k^T A V_k in Algorithm 3

	double eta;//c_1 in the paper, needed for LCOMM_OPT in the subproblem solve
	int l;
	int n;
	int Plength;//Number of columns in P, namely m in the paper
	int wcounter;//pointer to the column in P for the latest w
	int gcounter;//pointer to the column in P for the latest gradient

	double *sqrtD_XTP;//Temp matrix for computing sqrt(D_w) * X^T P for further forming the linear system
private:
	void update_inner_products(double *inner_product_matrix, double *P, double *t);// Update the matrix P^T A P stored in inner_product_matrix. t is the coefficients for forming the update direction (minimizer of (3))
	void compute_system(double *inner_product_matrix, double *XTP, double *sqrt_C_times_D, double *system);// Compute the coefficients in the linear system; double *system stores the output
	void compute_system(double *inner_product_matrix, double *XTP, double *sqrt_C_times_D, double *system, int *index, int index_length);// Compute the coefficients in the linear system when D_w is sparse, with the index set given in *index and index_length is the size of the index set
	int check_stopping(double *XTP, double *gcoeff, double *accumulated_t, double *inner_product_matrix);// Check if the stopping condition (21) for the inner iterations is satisfied. Return 1 when satisfied and 0 otherwise. (P * accumulated_t) forms the update direction G(t)
	int multi_iter;// 1 if run multiple inner iterations and 0 otherwise
	double v_coeff[3];//The coefficients in combining the columns of P to update v; see eq. (7) in the paper
	int vcounter;//Pointing to the column in P that stores the vector v; This is always set to 2 throughout
private:
	double sigma;
};

class AG: public SOLVER // Accelerated gradient of Nesterov. Implemented for comparison
{
public:
	AG(const function *fun_obj, double eps = 0.1, int linesearch = 1, int max_iter = 500000);

	void solve(double *w);
private:
	int linesearch;
};


class LCOMM_BFGS: public LCOMM_OPT
{
public:
	LCOMM_BFGS(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 500000);
	~LCOMM_BFGS();

	void solve(double *w);
protected:
	int tmp_length;// Number of columns in the current P. Upper bounded by Plength
	int M; //Number of previous iterations whose historical information we use to form P. Plength = 2*M + 2.

private:
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size);// Update the matrix P^T A P stored in inner_product_matrix. t is the coefficients for forming the update direction (minimizer of (3)). DynamicM is the current number of historical iterations used (DynamicM = min(M, iter)). previous_M is the value of DynamicM in last iteration
	 
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse);// Compute the coefficients in the linear system; double *system stores the output. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length);// Compute the coefficients in the linear system when D_w is sparse, with the index set given in *index and index_length is the size of the index set. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns

};

class LCOMM_DIAG: public LCOMM_BFGS
{
public:
	LCOMM_DIAG(const function *fun_obj, int l, double eps = 0.1, int m=10, double eta = 1e-2, int max_iter = 500000);
	~LCOMM_DIAG();

	void solve(double *w);
private:
	int dcounter;//The column in P that maintains the current D^{-1} grad f
	void update_inner_products(double* inner_product_matrix, int DynamicM, int previous_M, double* P, double* t, double step_size);// Update the matrix P^T A P stored in inner_product_matrix. t is the coefficients for forming the update direction (minimizer of (3)). DynamicM is the current number of historical iterations used (DynamicM = min(M, iter)). previous_M is the value of DynamicM in last iteration
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse);// Compute the coefficients in the linear system; double *system stores the output. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
	void compute_system(double* inner_product_matrix, double* XTP, double* sqrt_C_times_D, double* system, int sparse, int *index, int index_length);// Compute the coefficients in the linear system when D_w is sparse, with the index set given in *index and index_length is the size of the index set. sparse = 1 if DynamicM < M, and in this case the columns storing w_i, i > 0 can be ignored in the linear system as they are linear combinations of other columns
};
#endif
